/** @file sp/spAudio.h
 */

#ifndef __SPAUDIO_H
#define __SPAUDIO_H

#if defined(_MSC_VER)
#pragma comment ( lib, "winmm" )
#endif

#include <sp/spDefs.h>
#include <sp/spOption.h>
#include <sp/spAudioDefs.h>

#if defined(SGI) || defined(linux) || defined(OSS) || defined(_WIN32) || defined(sun) || defined(MACOS) || defined(ANDROID)
#define SP_SUPPORT_AUDIO
#endif

/*
#if defined(MACOSX64)
#undef SP_SUPPORT_AUDIO
#endif
*/

#ifdef __cplusplus
extern "C" {
#endif

/** @~english @defgroup spAudioGroup Audio I/O */

/** @addtogroup spAudioGroup
 *  @{ */  /*----@addtogroup spAudioGroup----*/

/** @~english
 *  @class spAudio spAudio.h <sp/spAudio.h>
 *  @brief A class to handle audio I/O.
@details Actually, spAudio is not a class, but a typedefed type by `typedef struct _spAudio *spAudio;`.
You can use this class after initialization by spInitAudio() or spInitAudioDriver() , should free resources of this class
after use by spFreeAudio() or spFreeAudioDriver() .

Normal audio I/O procedure is as follows.
  -# Get spAudio by spInitAudio() or spInitAudioDriver() .
  -# Select an audio device by spSelectAudioDevice() .
     If you don't call this function, the default device will be used.
  -# Set information of the current device by spSetAudioXXX
  -# Open the current device by spOpenAudioDevice()
  -# Check the device information by spGetAudioXXX if necessary since the current information is
     sometimes not indentical to the information you set.
  -# Call spReadAudio() for input or spWriteAudio() for output repeatedly.
  -# Flush buffers by spSyncAudio() if you want to restart I/O.
  -# Close the current device by spCloseAudioDevice()
  -# Free resources of spAudio by spFreeAudio() or spFreeAudioDriver() .
  .
 *  @sa spInitAudio() , spInitAudioDriver() , spFreeAudio() , spFreeAudioDriver()
 */
typedef struct _spAudio *spAudio;  /**< @~english Definition of spAudio. */

#define SP_AUDIO_BLOCKING 0     /**< @~english Constant to enable blocking mode for spSetAudioBlockingMode() .  */
#define SP_AUDIO_NONBLOCKING 1  /**< @~english Constant to disable blocking mode for spSetAudioBlockingMode() .  */

#define SP_AUDIO_CAPS_OPTION_DIALOG (1<<5)    
    
typedef unsigned long spAudioCallbackType;  /**< @~english Type used for spSetAudioCallbackFunc() . Must be a combination of SP_AUDIO_*_CALLBACK. */
    
#define SP_AUDIO_NO_CALLBACK 0  /**< @~english No callback exists. */
#define SP_AUDIO_OUTPUT_POSITION_CALLBACK (1<<0)  /**< @~english Called when the playback position changed.@n data1: spLong *position, data2: NULL */
#define SP_AUDIO_OUTPUT_BUFFER_CALLBACK (1<<2)  /**< @~english Called when the current buffer finished.@n data1: void *buffer, data2: long *buffer_length */

#if 0
/* reserved */
#define SP_AUDIO_OUTPUT_FINISHED_CALLBACK (1<<3)  /* data1: void *buffer, data2: long *buffer_length */
#define SP_AUDIO_OUTPUT_BUFFER_REQUEST_CALLBACK (1<<4)  /* data1: void *buffer, data2: long *buffer_length */
#define SP_AUDIO_INPUT_POSITION_CALLBACK (1<<5)  /* data1: void *buffer, data2: long *buffer_length */
#define SP_AUDIO_INPUT_BUFFER_CALLBACK (1<<6)  /* data1: void *buffer, data2: long *buffer_length */
#define SP_AUDIO_INPUT_FINISHED_CALLBACK (1<<7)  /* data1: void *buffer, data2: long *buffer_length */
#define SP_AUDIO_INPUT_BUFFER_REQUEST_CALLBACK (1<<8)  /* data1: void *buffer, data2: long *buffer_length */
#endif
    
/** @~english Signature of the callback function for spSetAudioCallbackFunc() . */
typedef spBool (*spAudioCallbackFunc)(spAudio audio, spAudioCallbackType call_type, void *data1, void *data2, void *user_data);
    
/** @} */  /*----@addtogroup spAudioGroup----*/

#if defined(MACOS)
#pragma import on
#endif

/** @addtogroup spAudioGroup
 *  @{ */  /*----@addtogroup spAudioGroup----*/

/** @~english Gets the number of audio drivers. */
extern int spGetNumAudioDriver(void);

/** @~english Gets the name of the audio driver identified by the index.
 *  @param[in] index 0-based index. You can specify an index from 0 to ( spGetNumAudioDriver() - 1).
 *  @return A null-terminated string indicating the name of the audio driver. You have to deallocate the memory by calling xspFree() . */
extern char *xspGetAudioDriverName(int index);
    
/** @~english Gets the number of devices for the audio driver.
 *  If `driver_name == NULL`, this function gets it for all drivers.
 *  @param[in] driver_name A null-terminated string of the driver name.
 *  @return The numer of the device; or -1 if failed. */
extern int spGetNumAudioDriverDevice(const char *driver_name);

/** @~english Gets the name of the audio device for the audio driver. 
 *  If `driver_name == NULL`, this function gets it for all drivers.
 *  @param[in] driver_name A null-terminated string of the driver name.
 *  @param[in] index 0-based index. You can specify an index from 0 to ( spGetNumAudioDriverDevice() - 1).
 *  @return A null-terminated string indicating the name of the device. You have to deallocate the memory by calling xspFree() . */
extern char *xspGetAudioDriverDeviceName(const char *driver_name, int index);
/** @} */  /*----@addtogroup spAudioGroup----*/
    
/** @~english @name Initialization
 *  @{ */

/** @~english Initializes the audio driver (constructor).
 *  @n All audio I/O functions do not work before you call this function (or spInitAudio() ).
 *  Multiple calls before spFreeAudioDriver() with one driver are not accepted. Some environments accept multiple calls of this function with different drivers.
 *  @param[in] driver_name The name of the driver.
 *  @sa spFreeAudioDriver()  @relatesalso spAudio */
extern spAudio spInitAudioDriver(const char *driver_name);
/** @~english Initializes the audio driver (constructor).
 *  @n All audio I/O functions do not work before you call this function (or spInitAudioDriver() ).
 *  Multiple calls before spFreeAudio() are not accepted. 
 *  @sa spFreeAudio()  @relatesalso spAudio */
extern spAudio spInitAudio(void);
/** @~english Frees the audio driver's resources. Use spFreeAudioDriver() instead of this.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() .
 *  @relatesalso spAudio */
extern void _spFreeAudioDriver(spAudio audio);
/** @internal Add a callback called at exit of the application.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spAddAudioExitCallback(spAudio audio, spExitCallbackFunc callback);
/** @~english @} */

/** @~english @name Device Selection
 *  @{ */
/** @~english Gets the number of audio devices.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] num_device The address of a variable to receive the number of devices.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetNumAudioDevice(spAudio audio, int *num_device);
/** @~english Gets the name of the audio device. You have to deallocate the memory by calling xspFree() .
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] device_index 0-based device index. You can specify an index from 0 to ( spGetNumAudioDevice() - 1).
 *  The index 0 always corresponds to the default device.
 *  @return A null-terminated string indicating the name of the device. You have to deallocate the memory by calling xspFree() .
 *  @relatesalso spAudio */
extern char *xspGetAudioDeviceName(spAudio audio, int device_index);
/** @~english Selects an audio device which has the index specified.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] device_index 0-based device index. You can specify an index from 0 to ( spGetNumAudioDevice() - 1).
 *  The index 0 always corresponds to the default device.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSelectAudioDevice(spAudio audio, int device_index);
/** @~english @} */

/** @internal
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spIsAudioOptionDialogExist(spAudio audio);
/** @internal
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spDisplayAudioOptionDialog(spAudio audio);

/** @name Setter and Getter
 *  @{ */
/** @~english Sets sample rate of the current device. This function should be called before opening a device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] samp_rate The value of sample rate in Hz.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSetAudioSampleRate(spAudio audio, double samp_rate);
/** @~english Gets sample rate of the current device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] samp_rate The address of a variable to receive the sample rate in Hz.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioSampleRate(spAudio audio, double *samp_rate);
/** @~english Sets the number of channels of the current device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] num_channel The number of channels.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSetAudioChannel(spAudio audio, int num_channel);
/** @~english Gets the number of channels of the current device. This function should be called before opening a device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] num_channel The address of a variable to receive the number of channels.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioChannel(spAudio audio, int *num_channel);
/** @~english Sets the buffer size of the current device. This function should be called before opening a device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] buffer_size The buffer size.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSetAudioBufferSize(spAudio audio, int buffer_size);
/** @~english Gets the buffer size of the current device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] buffer_size The address of a variable to receive the buffer size.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioBufferSize(spAudio audio, int *buffer_size);
/** @~english Sets the number of buffers of the current device. This function should be called before opening a device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] num_buffer The number of buffers.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSetAudioNumBuffer(spAudio audio, int num_buffer);
/** @~english Gets the number of buffers of the current device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] num_buffer The address of a variable to receive the number of buffers.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioNumBuffer(spAudio audio, int *num_buffer);
/** @~english Sets the blocking mode of the current device. This function should be called before opening a device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] block_mode The blocking mode of #SP_AUDIO_BLOCKING or #SP_AUDIO_NONBLOCKING .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSetAudioBlockingMode(spAudio audio, int block_mode);
/** @~english Gets the blocking mode of the current device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] block_mode The address of a variable to receive the blocking mode.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioBlockingMode(spAudio audio, int *block_mode);
/** @~english Sets bits/sample of the current device. This function should be called before opening a device.
 *  16, 24 and 32 bits/sample are supported currently. For the device not supporting 24 and 32 bits/sample, the data is internally converted to 16 bits/sample data. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] samp_bit The bits/sample.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSetAudioSampleBit(spAudio audio, int samp_bit);
/** @~english Gets bits/sample of the current device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] samp_bit The address of a variable to receive the bits/sample.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioSampleBit(spAudio audio, int *samp_bit);
/** @~english Gets the bits/sample specified.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] specified_samp_bit The address of a variable to receive the bits/sample specified.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioSpecifiedSampleBit(spAudio audio, int *specified_samp_bit);
/** @~english Sets a callback function. This function should be called before opening a device.
 *  The callback types of #SP_AUDIO_OUTPUT_POSITION_CALLBACK and #SP_AUDIO_OUTPUT_BUFFER_CALLBACK are only supported currently.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] call_type The callback type. Multiple types can be specified by using '|'.
 *  @param[in] call_func The callback function which has the following signature ( #spAudioCallbackFunc ): @code
      spBool func(spAudio audio, spAudioCallbackType call_type, void *data1, void *data2, void *user_data);
 *  @endcode
 *  @param[in] call_data The user data for the callback function. You can obtain this address as the 5th argument of the callback function.
 *  @return The callack type which can be set successfully. This value sometimes differs from \p call_type value on some environments.
 *  @relatesalso spAudio */
extern spAudioCallbackType spSetAudioCallbackFunc(spAudio audio, spAudioCallbackType call_type,
                                                  spAudioCallbackFunc call_func, void *call_data);
/** @} */

/** @~english @name Open and Close
 *  @{ */
/** @~english Opens the current audio device. If you want full-duplex I/O, open with read mode then do with write mode.
 *  The second open with write mode will fail if the device doesn't accept full-duplex I/O.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] mode Device opening mode. 'r' means read mode, 'w' means write mode. 'ro' and 'wo' which mean read only
 *  and write only modes are also supported. Although these `only' modes invalidate full-duplex I/O, sometimes improve sound quality especially for Linux OSS environment.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spOpenAudioDevice(spAudio audio, const char *mode);
/** @~english Closes the current audio device.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spCloseAudioDevice(spAudio audio);
/** @~english @} */

/** @~english @name Read and Write
 *  @{ */
/** @~english Reads raw data from the audio device. This function blocks processing until the read length reaches
 *  \p length unless the blocking mode is disabled by spSetAudioBlockingMode().
 *  @warning This function can be used when spOpenAudioDevice() is called with `"r"` mode. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] data The address of a buffer to receive raw data from the audio device. The type of the buffer must be 'short' for up to 16 bits/sample, 'long' for up to 32 bits/sample.
 *  @param[in] length The length of the buffer. This length is not the size in bytes of the buffer, but the number of elements of the buffer.
 *  @return The read length for success, or -1 for failure.
 *  @relatesalso spAudio */
extern long spReadAudio(spAudio audio, void *data, long length);
extern long spReadAudioBuffer(spAudio audio, char *buffer, long size);
/** @~english Writes raw data to the audio device. Since writing data is performed asynchronously, the function call finishes before playing whole specified data. The timing of the finish depends on your environment.
 *  @warning This function can be used when spOpenAudioDevice() is called with `"w"` mode. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] data The address of a buffer to send raw data to the audio device. The type of the buffer must be 'short' for up to 16 bits/sample, 'long' for up to 32 bits/sample.
 *  @param[in] length The length of the buffer. This length is not the size in bytes of the buffer, but the number of elements of the buffer.
 *  @return The written length for success, -1 for failure.
 *  @relatesalso spAudio */
extern long spWriteAudio(spAudio audio, void *data, long length);
extern long spWriteAudioBuffer(spAudio audio, char *buffer, long size);
/** @~english Reads data to a double array from the audio device. The \p weight is multiplied to the data after reading.
 *  @warning This function can be used when spOpenAudioDevice() is called with `"r"` mode. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] data The address of a double array to receive data from the audio device.
 *  @param[in] length The length of the buffer. This length is not the size in bytes of the buffer, but the number of elements of the buffer.
 *  @param[in] weight a weighting factor multiplied to each element in \p data after reading.
 *  @return The read length for success, -1 for failure.
 *  @relatesalso spAudio */
extern long spReadAudioDoubleWeighted(spAudio audio, double *data, long length, double weight);
/** @~english Reads data to a double array from the audio device.
 *  @warning This function can be used when spOpenAudioDevice() is called with `"r"` mode. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] data The address of a double array to receive data from the audio device.
 *  @param[in] length The length of the buffer. This length is not the size in bytes of the buffer, but the number of elements of the buffer.
 *  @return The read length for success, -1 for failure.
 *  @relatesalso spAudio */
extern long spReadAudioDouble(spAudio audio, double *data, long length);
/** @~english Writes a double array to the audio device. The \p weight is multiplied to the data before writing.
 *  @warning This function can be used when spOpenAudioDevice() is called with `"w"` mode. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] data The address of a double array to send data to the audio device. The data range must be basically from -1 to 1 inclusive.
 *  @param[in] length The length of the buffer. This length is not the size in bytes of the buffer, but the number of elements of the buffer.
 *  @param[in] weight a weighting factor multiplied to each element in \p data before writing.
 *  @return The written length for success, -1 for failure.
 *  @relatesalso spAudio */
extern long spWriteAudioDoubleWeighted(spAudio audio, double *data, long length, double weight);
/** @~english Writes a double array to the audio device.
 *  @warning This function can be used when spOpenAudioDevice() is called with `"w"` mode. 
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[in] data The address of a double array to send data to the audio device. The data range must be basically from -1 to 1 inclusive.
 *  @param[in] length The length of the buffer. This length is not the size in bytes of the buffer, but the number of elements of the buffer.
 *  @return The written length for success, -1 for failure.
 *  @relatesalso spAudio */
extern long spWriteAudioDouble(spAudio audio, double *data, long length);
/** @~english @} */
    
/** @~english @name Device Control
 *  @{ */
/** @~english Gets the playback position.
 *  @note Since this function is not supported on some environments, the function call on those environments always fails.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @param[out] position The address of a variable to receive the playback position.
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spGetAudioOutputPosition(spAudio audio, spLong *position);
/** @~english Stops audio I/O as soon as possible. Note that the behavior of this function has dependence on an environment.
 *  The function call on some environments will decrease a processing load starting to increase from beginning of audio I/O.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spStopAudio(spAudio audio);
/** @~english Synchronizes audio I/O. Note that the behavior of this function has dependence on an environment.
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() or spInitAudio() .
 *  @retval SP_TRUE Success  @retval SP_FALSE Failure
 *  @relatesalso spAudio */
extern spBool spSyncAudio(spAudio audio);
/** @~english @} */
    
#if defined(MACOS)
#pragma import off
#endif

/** @addtogroup spAudioGroup
 *  @{ */
/** @~english Frees the audio driver's resources (destructor).
 *  @param[in] audio spAudio type variable obtained by spInitAudioDriver() . */
#define spFreeAudioDriver(audio) {_spFreeAudioDriver(audio); (audio) = NULL;}
/** @~english Frees the default audio driver's resources (destructor).
 *  @param[in] audio spAudio type variable obtained by spInitAudio() .
 *  @relatesalso spAudio */
#define spFreeAudio(audio) spFreeAudioDriver(audio)
/** @} */

#ifdef __cplusplus
}  /* Close scope of 'extern "C"' declaration */
#endif

#endif /* __SPAUDIO_H */
