/** @file sp/spWave.h
 */

#ifndef __SPWAVE_H
#define __SPWAVE_H

#include <sp/spDefs.h>
#include <sp/spAudioDefs.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @~english @defgroup spWaveGroup Audio File */

/** @addtogroup spWaveGroup
 *  @{ */

#define SP_WAVE_WAV_ID "wav"
#define SP_WAVE_AIFF_ID "aiff"
#define SP_WAVE_AIFC_ID "aifc"
    
#if BYTE_ORDER == BIG_ENDIAN
#define SP_WAV_NEED_SWAP 1
#else
#define SP_WAV_NEED_SWAP 0
#endif

#if BYTE_ORDER == BIG_ENDIAN
#define SP_AIFF_NEED_SWAP 0
#else
#define SP_AIFF_NEED_SWAP 1
#endif

#define SP_WAVE_FORMAT_SIZE 32
#define SP_WAVE_FILE_TYPE_SIZE 32       /**< Size of `file_type` member of spWaveInfo . */
#define SP_WAVE_FILE_DESC_SIZE 128      /**< Size of `file_desc` member of spWaveInfo . */
#define SP_WAVE_FILE_FILTER_SIZE 32     /**< Size of `file_filter` member of spWaveInfo . */

/** @~english
 *  @struct spWaveInfo spWave.h <sp/spWave.h>
 *  A structure to hold information related to the audio file. */
typedef struct _spWaveInfo {
    char file_type[SP_WAVE_FILE_TYPE_SIZE];     /**< @~english Format unique ID, e.g. `"wav"` . */
    char file_desc[SP_WAVE_FILE_DESC_SIZE];     /**< @~english Format description, e.g. `"Microsoft PCM"` . */
    char file_filter[SP_WAVE_FILE_FILTER_SIZE]; /**< @~english Filter mask, e.g. `"*.wav"` . */
    int buffer_size;            /**< @~english Buffer size for output buffer. */
    spLong header_size;         /**< @~english You can skip header by fseek with this size. If you aren't allowed that, zero will be set. */
    int samp_bit;               /**< @~english Bits/sample. */
    int num_channel;            /**< @~english The number of channels. */
    double samp_rate;           /**< @~english Sampling rate [Hz]. */
    long bit_rate;              /**< @~english Bit rate [bits/sec]. */
    spLong length;              /**< @~english Total length of sound [point]. */
} spWaveInfo;

#define SP_SONG_NO_INFO 0       /**< Mask to indicate no information. */
#define SP_SONG_TRACK_MASK (1L<<0) /**< Mask for `track` . */
#define SP_SONG_TITLE_MASK (1L<<1) /**< Mask for `title` . */
#define SP_SONG_ARTIST_MASK (1L<<2) /**< Mask for `artist` . */
#define SP_SONG_ALBUM_MASK (1L<<3)  /**< Mask for `album` . */
#define SP_SONG_GENRE_MASK (1L<<4)  /**< Mask for `genre` . */
#define SP_SONG_RELEASE_MASK (1L<<5) /**< Mask for `release` . */
#define SP_SONG_COPYRIGHT_MASK (1L<<6) /**< Mask for `copyright` . */
#define SP_SONG_ENGINEER_MASK (1L<<7)  /**< Mask for `engineer` . */
#define SP_SONG_SOURCE_MASK (1L<<8)    /**< Mask for `source` . */
#define SP_SONG_SOFTWARE_MASK (1L<<9)  /**< Mask for `software` . */
#define SP_SONG_SUBJECT_MASK (1L<<10)  /**< Mask for `subject` . */
#define SP_SONG_COMMENT_MASK (1L<<11)  /**< Mask for `comment` . */
#define SP_SONG_ALBUM_ARTIST_MASK (1L<<12) /**< Mask for `album_artist` . V2 only. */
#define SP_SONG_COMPOSER_MASK (1L<<13)     /**< Mask for `composer` . V2 only. */
#define SP_SONG_LYRICIST_MASK (1L<<14)     /**< Mask for `lyricist` . V2 only. */
#define SP_SONG_PRODUCER_MASK (1L<<15)     /**< Mask for `producer` . V2 only. */
#define SP_SONG_ISRC_MASK (1L<<16)         /**< Mask for `isrc` . V2 only. */
#define SP_SONG_TRACK_TOTAL_MASK (1L<<17)  /**< Mask for `track_total` . V2 only. */
#define SP_SONG_DISC_MASK (1L<<18)         /**< Mask for `disc` . V2 only. */
#define SP_SONG_DISC_TOTAL_MASK (1L<<19)   /**< Mask for `disc_total` . V2 only. */
#define SP_SONG_TEMPO_MASK (1L<<20)        /**< Mask for `tempo` . V2 only. */
#define SP_SONG_ID3V1_MASK (1L<<24)        /**< Mask to indicate ID3v1. */
#define SP_SONG_ID3V2_2_MASK (1L<<25)      /**< Mask to indicate ID3v2.2. */
#define SP_SONG_ID3V2_3_MASK (1L<<26)      /**< Mask to indicate ID3v2.3. */
#define SP_SONG_ID3V2_4_MASK (1L<<27)      /**< Mask to indicate ID3v2.4. */
#define SP_SONG_UTF8_STRING_IN_FILE_MASK (1L<<28)  /**< Mask to indicate that the format of strings in file is UTF8. */
#define SP_SONG_ISO8859_1_STRING_IN_FILE_MASK (1L<<29)  /**< Mask to indicate that the format of strings in file is ISO-8859-1. */

/** All masks supported by ID3v2. */
#define SP_SONG_ID3V2_SUPPORT_MASK (SP_SONG_TRACK_MASK|SP_SONG_TITLE_MASK|SP_SONG_ARTIST_MASK|SP_SONG_ALBUM_MASK|SP_SONG_GENRE_MASK|SP_SONG_RELEASE_MASK|SP_SONG_COPYRIGHT_MASK|SP_SONG_ENGINEER_MASK|SP_SONG_SOURCE_MASK|SP_SONG_SOFTWARE_MASK|SP_SONG_COMMENT_MASK|SP_SONG_ALBUM_ARTIST_MASK|SP_SONG_COMPOSER_MASK|SP_SONG_LYRICIST_MASK|SP_SONG_ISRC_MASK|SP_SONG_TRACK_TOTAL_MASK|SP_SONG_DISC_MASK|SP_SONG_DISC_TOTAL_MASK|SP_SONG_TEMPO_MASK)
/** All masks to indicate that ID3v2.2 is supported. */
#define SP_SONG_ID3V2_2_SUPPORT_MASK (SP_SONG_ID3V2_SUPPORT_MASK|SP_SONG_ID3V2_2_MASK)
/** All masks to indicate that ID3v2.3 is supported. */
#define SP_SONG_ID3V2_3_SUPPORT_MASK (SP_SONG_ID3V2_SUPPORT_MASK|SP_SONG_ID3V2_3_MASK)
/** All masks to indicate that ID3v2.4 is supported. */
#define SP_SONG_ID3V2_4_SUPPORT_MASK (SP_SONG_ID3V2_SUPPORT_MASK|SP_SONG_ID3V2_4_MASK)
/** All masks to indicate that all ID3v2 versions are supported. */
#define SP_SONG_ID3V2_ALL_SUPPORT_MASK (SP_SONG_ID3V2_SUPPORT_MASK|SP_SONG_ID3V2_2_MASK|SP_SONG_ID3V2_3_MASK|SP_SONG_ID3V2_4_MASK)
    
#define SP_SONG_INFO_SIZE 128   /**< Size of string members of spSongInfo . */
    
#define SP_SONG_INFO_GENRE_SIZE 32 /**< Size of `genre` of spSongInfoV2 .  */
#define SP_SONG_INFO_ALBUM_ARTIST_SIZE 96 /**< Size of `album_artist` of spSongInfoV2 .  */
    
#define SP_SONG_INFO_RELEASE_SIZE 48 /**< Size of `release` of spSongInfoV2 .  */
#define SP_SONG_INFO_PRODUCER_SIZE 64 /**< Size of `producer` of spSongInfoV2 .  */
    
#define SP_SONG_INFO_SOURCE_SIZE 48 /**< Size of `source` of spSongInfoV2 .  */
#define SP_SONG_INFO_COMPOSER_SIZE 80 /**< Size of `composer` of spSongInfoV2 .  */
    
#define SP_SONG_INFO_SOFTWARE_SIZE 48 /**< Size of `software` of spSongInfoV2 .  */
#define SP_SONG_INFO_LYRICIST_SIZE 80 /**< Size of `lyricist` of spSongInfoV2 .  */
    
#define SP_SONG_INFO_SUBJECT_SIZE 104 /**< Size of `subject` of spSongInfoV2 .  */
#define SP_SONG_INFO_ISRC_SIZE 24 /**< Size of `isrc` of spSongInfoV2 .  */

/** @~english
 *  @struct spSongInfo spWave.h <sp/spWave.h>
 *  A structure to hold song information.
 *  @sa spSongInfoV2 */
typedef struct _spSongInfo {
    unsigned long info_mask;            /**< @~english Mask indicating which information is valid. */
    spLong32 track;                     /**< @~english Track number. */
    char title[SP_SONG_INFO_SIZE];      /**< @~english Song title. */
    char artist[SP_SONG_INFO_SIZE];     /**< @~english Artist name. */
    char album[SP_SONG_INFO_SIZE];      /**< @~english Album title. */
    char genre[SP_SONG_INFO_SIZE];      /**< @~english Genre. */
    char release[SP_SONG_INFO_SIZE];    /**< @~english Release date. */
    char copyright[SP_SONG_INFO_SIZE];  /**< @~english Copyright information. */
    char engineer[SP_SONG_INFO_SIZE];   /**< @~english Engineer name. */
    char source[SP_SONG_INFO_SIZE];     /**< @~english Source information. WAVE file uses this information. */
    char software[SP_SONG_INFO_SIZE];   /**< @~english Software name. */
    char subject[SP_SONG_INFO_SIZE];    /**< @~english Subject. */
    char comment[SP_SONG_INFO_SIZE];    /**< @~english Some comments. */
} spSongInfo;

/* 0.7.12+ */
/** @~english
 *  @struct spSongInfoV2 spWave.h <sp/spWave.h>
 *  A structure to hold song information (version 2). This is the extended version of spSongInfo .
 *  @sa spSongInfo */
typedef struct _spSongInfoV2 {
    unsigned long info_mask;                    /**< @~english Mask indicating which information is valid. */
    spLong32 track;                             /**< @~english Track number. */
    char title[SP_SONG_INFO_SIZE];              /**< @~english Song title. */
    char artist[SP_SONG_INFO_SIZE];             /**< @~english Artist name. */
    char album[SP_SONG_INFO_SIZE];              /**< @~english Album title. */
    char genre[SP_SONG_INFO_GENRE_SIZE];        /**< @~english Genre. */
    char album_artist[SP_SONG_INFO_ALBUM_ARTIST_SIZE]; /**< @~english Album artist. V2 only. */
    char release[SP_SONG_INFO_RELEASE_SIZE];    /**< @~english Release date. */
    char producer[SP_SONG_INFO_PRODUCER_SIZE];  /**< @~english Producer name. V2 only. */
    spLong32 track_total;                       /**< @~english The number of total tracks. V2 only. */
    spLong32 disc;                              /**< @~english Disc number. V2 only. */
    spLong32 disc_total;                        /**< @~english The numer of total discs. V2 only. */
    spLong32 tempo;                             /**< @~english Tempo in BPM. V2 only. */
    char copyright[SP_SONG_INFO_SIZE];          /**< @~english Copyright information. */
    char engineer[SP_SONG_INFO_SIZE];           /**< @~english Engineer name. */
    char source[SP_SONG_INFO_SOURCE_SIZE];      /**< @~english Source information. WAVE file uses this information. */
    char composer[SP_SONG_INFO_COMPOSER_SIZE];  /**< @~english Composer name. V2 only. */
    char software[SP_SONG_INFO_SOFTWARE_SIZE];  /**< @~english Software name. */
    char lyricist[SP_SONG_INFO_LYRICIST_SIZE];  /**< @~english Lyricist name. V2 only. */
    char subject[SP_SONG_INFO_SUBJECT_SIZE];    /**< @~english Subject. */
    char isrc[SP_SONG_INFO_ISRC_SIZE];          /**< @~english ISRC information. V2 only. */
    char comment[SP_SONG_INFO_SIZE];            /**< @~english Some comments. */
} spSongInfoV2;
    
typedef struct _spSongInfo *spSongInfoPtr;
typedef struct _spSongInfoV2 *spSongInfoV2Ptr;
    
/** @} */

#if defined(MACOS)
#pragma import on
#endif

extern spBool spIsWavFile(const char *filename);
extern spBool spReadWavInfo(spWaveInfo *wave_info, FILE *fp);
extern spBool spWriteWavInfo(spWaveInfo *wave_info, FILE *fp);
extern long _spReadWavData(spWaveInfo *wave_info, char *data, long length, FILE *fp);
extern long _spWriteWavData(spWaveInfo *wave_info, char *data, long length, FILE *fp);
extern unsigned long spGetWavSongInfoMask(void);
extern spBool spGetWavSongInfo(spSongInfo *song_info, FILE *fp); /* fp must be located at file start or data end */
extern spBool spReadWavSongInfo(spWaveInfo *wave_info, spSongInfo *song_info, FILE *fp);
extern spBool spAddWavSongInfo(spSongInfo *song_info, FILE *fp); /* fp must be located at file start or data end */
extern spBool spWriteWavSongInfo(spWaveInfo *wave_info, spSongInfo *song_info, FILE *fp);

/* (data buffer size) = length * sizeof(short), (samp_bit <= 16)
   (data buffer size) = length * sizeof(long), (16 < samp_bit <= 32) */
#define spReadWavData(wave_info, data, length, fp) _spReadWavData(wave_info, (char *)(data), length, fp)
#define spWriteWavData(wave_info, data, length, fp) _spWriteWavData(wave_info, (char *)(data), length, fp)

extern spBool spIsAiffFile(const char *filename);
extern spBool spReadAiffInfo(spWaveInfo *wave_info, FILE *fp);
extern spBool spWriteAiffInfo(spWaveInfo *wave_info, FILE *fp);
extern long _spReadAiffData(spWaveInfo *wave_info, char *data, long length, FILE *fp);
extern long _spWriteAiffData(spWaveInfo *wave_info, char *data, long length, FILE *fp);
extern unsigned long spGetAiffSongInfoMask(void);
extern spBool spGetAiffSongInfo(spSongInfo *song_info, FILE *fp); /* location of fp is not restricted */
extern spBool spReadAiffSongInfo(spWaveInfo *wave_info, spSongInfo *song_info, FILE *fp);
extern spBool spAddAiffSongInfo(spSongInfo *song_info, FILE *fp); /* fp must be located at file start or data end */
extern spBool spWriteAiffSongInfo(spWaveInfo *wave_info, spSongInfo *song_info, FILE *fp);

#define spReadAiffData(wave_info, data, length, fp) _spReadAiffData(wave_info, (char *)(data), length, fp)
#define spWriteAiffData(wave_info, data, length, fp) _spWriteAiffData(wave_info, (char *)(data), length, fp)

/** @~english Initialize contents of spWaveInfo .  @relatesalso spWaveInfo */
extern spBool spInitWaveInfo(spWaveInfo *wave_info);
/** @~english Copy contents of spWaveInfo to another.  @relatesalso spWaveInfo */
extern spBool spCopyWaveInfo(spWaveInfo *dest_info, spWaveInfo *src_info);
/** @~english Initialize contents of spSongInfo .  @relatesalso spSongInfo */
extern spBool spInitSongInfo(spSongInfo *song_info);
/** @~english Initialize contents of spSongInfoV2 .  @relatesalso spSongInfoV2 */
extern spBool spInitSongInfoV2(spSongInfoV2 *song_info);
/** @~english Copy contents of spSongInfo to another.  @relatesalso spSongInfo */
extern spBool spCopySongInfo(spSongInfo *dest_info, spSongInfo *src_info);
/** @~english Copy contents of spSongInfoV2 to another.  @relatesalso spSongInfoV2 */
extern spBool spCopySongInfoV2(spSongInfoV2 *dest_info, spSongInfoV2 *src_info);
/** @~english Check if contents of two spSongInfo variables are identical.  @relatesalso spSongInfo */
extern spBool spEqSongInfo(spSongInfo *info1, spSongInfo *info2);
/** @~english Check if contents of two spSongInfoV2 variables are identical.  @relatesalso spSongInfoV2 */
extern spBool spEqSongInfoV2(spSongInfoV2 *info1, spSongInfoV2 *info2);
/** @~english Display contents of spSongInfo information.  @relatesalso spSongInfo */
extern spBool spShowSongInfo(spSongInfo *song_info, const char *indent, FILE *fp);
/** @~english Display contents of spSongInfoV2 information.  @relatesalso spSongInfoV2 */
extern spBool spShowSongInfoV2(spSongInfoV2 *song_info, const char *indent, FILE *fp);

/* the following functions may remove in future release */
extern void spPlayFileUseDatLink(spBool flag);
extern void spPlayFileUseWav(spBool flag);
extern spBool spSetPlayCommand(const char *format);
extern spBool spPlayFile(const char *filename, int num_channel, double samp_rate);
extern spBool spPlayFileAsync(const char *filename, int num_channel, double samp_rate);

#if defined(MACOS)
#pragma import off
#endif

#ifdef __cplusplus
}  /* Close scope of 'extern "C"' declaration */
#endif

#endif /* __SPWAVE_H */
