/*
 *        cafinfo.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spBaseLib.h>
#include <sp/spWave.h>
#include <sp/spWave_Caf.h>
#include <sp/spMain.h>

#define DEBUG

static spBool display_song_info;
static spBool help_flag;
static int debug_level = -1;

static spOptions options;
static spOption option[] = {
    {"-song", NULL, "display song information", NULL,
         SP_TYPE_BOOLEAN, &display_song_info, SP_FALSE_STRING},
    {"-debug", NULL, "debug level", NULL,
         SP_TYPE_INT, &debug_level, NULL},
    {"-help", "-h", "display this message", NULL,
         SP_TYPE_BOOLEAN, &help_flag, SP_FALSE_STRING},
};

static const char *filelabel[] = {
    "<filename...>",
};

static void showCafChunkInfo(spCafChunk *chunk, long depth)
{
    long k, l;
    spCafChunk *child_chunk;
    char type_string[5];
    spChunkSize chunk_size;
    spChunkSize content_size;
    spChunkSize margin_size;

    chunk_size = spGetCafChunkSize(chunk, type_string);
    content_size = spGetCafChunkContentSize(chunk, NULL);

    if (chunk_size > 0) {
        margin_size = spGetCafChunkContentMarginSize(chunk);
        
        for (k = 1; k <= depth; k++) printf("  ");
        if (margin_size > 0) {
            printf("'%s' (size: %ld (%ld + %ld) bytes, margin: %ld bytes)\n",
                   type_string, (long)chunk_size, (long)(chunk_size - content_size), (long)content_size, (long)margin_size);
        } else {
            printf("'%s' (size: %ld (%ld + %ld) bytes)\n", 
                   type_string, (long)chunk_size, (long)(chunk_size - content_size), (long)content_size);
        }

        if (strneq(type_string, "info", 4)) {
            char buf[SP_MAX_LINE];
            spCafInformationChunk *info = (spCafInformationChunk *)chunk;

            for (l = 0; l < info->mNumEntries; l++) {
                spStrCopy(buf, sizeof(buf), info->mInformations[l].mValue);
                spConvertKanjiToLocaleCode((unsigned char *)buf, sizeof(buf), SP_KANJI_CODE_UTF8);
                for (k = 1; k <= depth; k++) printf("  ");
                printf("  '%s': '%s'\n", info->mInformations[l].mKey, buf);
            }
        } else if (strneq(type_string, "desc", 4)) {
            spCafAudioDescChunk *desc = (spCafAudioDescChunk *)chunk;
            
            for (k = 1; k <= depth; k++) printf("  ");
            printf("  Format ID: '%c%c%c%c'\n", desc->mFormatID[0], desc->mFormatID[1], desc->mFormatID[2], desc->mFormatID[3]);

            for (k = 1; k <= depth; k++) printf("  ");
            printf("    sample rate = %.0f channels/frame = %ld, bits/channel = %ld\n",
                   desc->mSampleRate, desc->mChannelsPerFrame, desc->mBitsPerChannel);
            for (k = 1; k <= depth; k++) printf("  ");
            printf("    flags = %lx, bytes/packet = %ld, frames/packet = %ld\n",
                   desc->mFormatFlags, desc->mBytesPerPacket, desc->mFramesPerPacket);
        } else if (strneq(type_string, "pakt", 4)) {
            spCafPacketTableChunk *pakt = (spCafPacketTableChunk *)chunk;

            for (k = 1; k <= depth; k++) printf("  ");
            printf("  packets = %ld, valid frames = %ld, priming frames = %ld, remainder frames = %ld\n",
                   (long)pakt->mNumberPackets, (long)pakt->mNumberValidFrames, pakt->mPrimingFrames, pakt->mRemainderFrames);
        }
    }
    
    child_chunk = chunk->child;
    while (child_chunk != NULL) {
        showCafChunkInfo(child_chunk, depth + 1);
        child_chunk = child_chunk->next;
    }

    return;
}

int spMain(int argc, char *argv[])
{
    const char *filename;
    FILE *fp;
    long file_count;
    spChunkSize nread;
    spChunkSize total_read;
    spCafHeader caf_header;
    spSongInfoV2 song_info;
    
    spSetHelpMessage(&help_flag, "Get CAF file information (sample program)");
    options = spGetOptions(argc, argv, option, filelabel);
    spGetOptionsValue(argc, argv, options);

    spSetDebugLevel(debug_level);
    
    filename = spGetFile(options);
    if (filename == NULL) {
        spError(1, "Not enough files.\n");
    }

    file_count = 0;
    do {
        /* open file */
        if (NULL == (fp = spOpenFile(filename, "rb"))) {
            return 1;
        }
        
        printf("%s:\n", filename);
        total_read = 0;
        
        /* read information */
        if ((nread = spReadCafHeader(&caf_header, fp)) <= 0) {
            printf("  ");
            printf("This file is not CAF file.\n");
        } else {
            total_read += nread;

            showCafChunkInfo((spCafChunk *)&caf_header, 0);

            if (display_song_info == SP_TRUE) {
                if (spParseCafSongInfo(&caf_header, &song_info)) {
                    printf("\n");
                    spShowSongInfoV2(&song_info, NULL, stdout);
                    printf("\n");
                }
            }
            
            spFreeCafHeader(&caf_header);
        }
    
        /* close file */
        spCloseFile(fp);
        
        ++file_count;
    } while ((filename = spGetFile(options)) != NULL);
    
    return 0;
}
