/*
 *        id3info.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spBaseLib.h>
#include <sp/spWave.h>

#include <sp/spID3.h>
#include <sp/spMain.h>

#define DEBUG

static spBool display_song_info;
static spBool display_output_header_info;
static spBool help_flag;
static long file_offset = 0L;
static int debug_level = -1;

static char lang_str[SP_MAX_PATHNAME] = "";
static char o_id3_file[SP_MAX_PATHNAME] = "";

static spOptions options;
static spOption option[] = {
    {"-song", NULL, "display song information", NULL,
         SP_TYPE_BOOLEAN, &display_song_info, SP_FALSE_STRING},
    {"-lang", NULL, "convert string based on the specified LANG string", NULL,
         SP_TYPE_STRING_A, lang_str, NULL},
    {"-offset", NULL, "file offset before ID3 tag", NULL,
         SP_TYPE_LONG, &file_offset, "0"},
    {"-O", NULL, "output file (binary ID3 content)", NULL,
         SP_TYPE_STRING_A, o_id3_file, NULL},
    {"-oh", NULL, "display output header information", NULL,
         SP_TYPE_BOOLEAN, &display_output_header_info, SP_FALSE_STRING},
    {"-debug", NULL, "debug level", NULL,
         SP_TYPE_INT, &debug_level, NULL},
    {"-help", "-h", "display this message", NULL,
         SP_TYPE_BOOLEAN, &help_flag, SP_FALSE_STRING},
};

static const char *filelabel[] = {
    "<filename...>",
};

int spMain(int argc, char *argv[])
{
    const char *filename;
    FILE *fp;
    long file_count;
    char buf[SP_MAX_LINE];
    spBool flag;
    spID3FrameSize nread;
    spID3FrameSize total_read;
    spID3Header id3_header;
    spID3Header o_id3_header;
    spSongInfoV2 song_info;
    spID3Converter id3conv = NULL;
    
    spSetHelpMessage(&help_flag, "Get ID3 file information (sample program)");
    options = spGetOptions(argc, argv, option, filelabel);
    spGetOptionsValue(argc, argv, options);

    spSetDebugLevel(debug_level);
    
    filename = spGetFile(options);
    if (filename == NULL) {
        spError(1, "Not enough files.\n");
    }

    if (!spStrNone(lang_str)) {
        id3conv =  spID3ConverterOpen(lang_str, 0L);
        spDebug(-10, "main", "lang_str = %s, id3conv = %ld\n", lang_str, (long)id3conv);
    }

    file_count = 0;
    do {
        /* open file */
        if (NULL == (fp = spOpenFile(filename, "rb"))) {
            spError(1, "Cannot open input file: %s\n", filename);
        }
        
        if (id3conv != NULL) {
            spStrCopy(buf, sizeof(buf), filename);
            spConvertKanjiFromLocaleCode(buf, sizeof(buf), SP_KANJI_CODE_UTF8);
            spID3ConverterConvertUTF8StringTo(id3conv, buf, sizeof(buf));
            printf("%s:\n", buf);
        } else {
            printf("%s:\n", filename);
        }
        total_read = 0;

        if (file_offset > 0) {
            spSeekFile(fp, file_offset, SEEK_CUR);
        }
        
        /* read information */
        if ((nread = spReadID3Header(&id3_header, fp)) <= 0) {
            printf("  ");
            printf("This file is not ID3 file.\n");
        } else {
            total_read += nread;
            
            if (id3conv != NULL) {
                spID3ConverterShowID3HeaderInfo(id3conv, &id3_header);
            } else {
                spShowID3HeaderInfo(&id3_header);
            }

            if (display_song_info == SP_TRUE) {
                if (id3conv != NULL) {
                    flag = spID3ConverterParseSongInfo(id3conv, &id3_header, &song_info);
                } else {
                    flag = spParseID3SongInfo(&id3_header, &song_info);
                }
                
                if (flag) {
                    printf("\n");
                    spShowSongInfoV2(&song_info, NULL, stdout);
                    printf("\n");
                }
            }
            
            if (file_count == 0) {
                spCopyID3Header(&o_id3_header, &id3_header);
            }
            
            spFreeID3Header(&id3_header);
        
            ++file_count;
        }
    
        /* close file */
        spCloseFile(fp);
    } while ((filename = spGetFile(options)) != NULL);

    if (!spStrNone(o_id3_file)) {
        /* open file */
        if (NULL == (fp = spOpenFile(o_id3_file, "wb"))) {
            spError(1, "Cannot open output file: %s\n", o_id3_file);
        } else {
            spID3FrameSize nwrite;
            
            if (display_output_header_info == SP_TRUE) {
                if (id3conv != NULL) {
                    spStrCopy(buf, sizeof(buf), o_id3_file);
                    spConvertKanjiFromLocaleCode(buf, sizeof(buf), SP_KANJI_CODE_UTF8);
                    spID3ConverterConvertUTF8StringTo(id3conv, buf, sizeof(buf));
                    printf("Output ID3 File: %s\n", buf);
                } else {
                    printf("Output ID3 File: %s:\n", o_id3_file);
                }

                if (id3conv != NULL) {
                    spID3ConverterShowID3HeaderInfo(id3conv, &o_id3_header);
                } else {
                    spShowID3HeaderInfo(&o_id3_header);
                }
            }

            nwrite = spWriteID3Header(&o_id3_header, fp);
            spDebug(-10, "main", "nwrite = %ld\n", (long)nwrite);

            spCloseFile(fp);
        }
    }

    if (id3conv != NULL) {
        spID3ConverterClose(id3conv);
    }
    
    return 0;
}
